/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package es.rickyepoderi.apacheds;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import javax.naming.ldap.InitialLdapContext;
import javax.naming.ldap.LdapContext;
import org.junit.Assert;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * <p>Ranged interceptor to emulate the behavior of AD. AD has a limit in
 * the number of attributes that return (15000 by default in MaxValRange). 
 * See this MS link for AD limits:</p>
 *
 * https://support.microsoft.com/en-us/help/315071/how-to-view-and-set-ldap-policy-in-active-directory-by-using-ntdsutil
 *
 * <p>And this other link to know how range attribute search works:</p>
 *
 * https://docs.microsoft.com/en-us/previous-versions/windows/desktop/ldap/searching-using-range-retrieval
 *
 * @author ricky
 */
public class RangedAttributeInterceptorTest {

    static private Server server;
    static private LdapContext ctx;

    @BeforeClass
    public static void before() throws Exception {
        // start the server
        server = new Server(10389);
        // create a context to search
        Properties env = new Properties();
        env.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
        env.put(Context.PROVIDER_URL, "ldap://localhost:10389");
        env.put(Context.SECURITY_PROTOCOL, "plain");
        env.put("com.sun.jndi.ldap.connect.pool", "false");
        env.put(Context.SECURITY_AUTHENTICATION, "simple");
        env.put(Context.SECURITY_PRINCIPAL, "uid=admin,ou=system");
        env.put(Context.SECURITY_CREDENTIALS, "secret");
        ctx = new InitialLdapContext(env, null);
    }

    @AfterClass
    public static void after() throws Exception {
        server.stop();
    }

    private static List<Map<String, List<String>>> search(String baseDn, String filter, int scope, String... attrsToReturn) throws NamingException {
        SearchControls sc = new SearchControls();
        sc.setSearchScope(scope);
        if (attrsToReturn != null && attrsToReturn.length > 0) {
            sc.setReturningAttributes(attrsToReturn);
        }
        List<Map<String, List<String>>> result = new ArrayList<>();
        NamingEnumeration<SearchResult> ne = ctx.search(baseDn, filter, sc);
        while (ne.hasMore()) {
            SearchResult sr = ne.next();
            Map<String, List<String>> entry = new HashMap<>();
            entry.put("dn", Collections.singletonList(sr.getNameInNamespace()));
            NamingEnumeration<? extends Attribute> attrs = sr.getAttributes().getAll();
            while (attrs.hasMore()) {
                Attribute attr = attrs.next();
                NamingEnumeration<?> values = attr.getAll();
                List<String> entryValue = new ArrayList<>();
                while (values.hasMore()) {
                    Object value = values.next();
                    entryValue.add(value.toString());
                }
                entry.put(attr.getID(), entryValue);
            }
            result.add(entry);
        }
        return result;
    }

    private void checkUsers(List<String> members, int min, int max) {
        for (int i = min; i <= max; i++) {
            Assert.assertTrue("user" + i + " expected but was: " + members.get(i - min), members.get(i - min).startsWith("uid=user" + i));
        }
    }

    @Test
    public void testRequestUniqueMemberNoRangeNoLimit() throws NamingException {
        // group2 has 3 uniquemember so should not be ranged by default
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group2", SearchControls.SUBTREE_SCOPE, "uniquemember");
        System.err.println(objs);
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember"));
        Assert.assertEquals(3, objs.get(0).get("uniquemember").size());
        checkUsers(objs.get(0).get("uniquemember"), 0, 2);
    }
    
    @Test
    public void testRequestUniqueMemberForceRange() throws NamingException {
        // group2 has 3 uniquemember but range requested
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group2", SearchControls.SUBTREE_SCOPE, "uniquemember;range=0-1");
        System.err.println(objs);
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=0-1"));
        Assert.assertEquals(2, objs.get(0).get("uniquemember;range=0-1").size());
        checkUsers(objs.get(0).get("uniquemember;range=0-1"), 0, 1);
    }
    
    @Test
    public void testRequestUniqueMemberForceRangeExact() throws NamingException {
        // group2 has 3 uniquemember with exact range
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group2", SearchControls.SUBTREE_SCOPE, "uniquemember;range=0-2");
        System.err.println(objs);
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=0-*"));
        Assert.assertEquals(3, objs.get(0).get("uniquemember;range=0-*").size());
        checkUsers(objs.get(0).get("uniquemember;range=0-*"), 0, 1);
    }
    
    @Test
    public void testRequestUniqueMemberForceRange0_() throws NamingException {
        // group2 has 3 uniquemember requesting all the range
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group2", SearchControls.SUBTREE_SCOPE, "uniquemember;range=0-*");
        System.err.println(objs);
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=0-*"));
        Assert.assertEquals(3, objs.get(0).get("uniquemember;range=0-*").size());
        checkUsers(objs.get(0).get("uniquemember;range=0-*"), 0, 1);
    }

    @Test
    public void testRequestUniqueMemberNoRangeLimit() throws NamingException {
        // group1 shoudl be ranged 11 members
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=0-4"));
        Assert.assertEquals(5, objs.get(0).get("uniquemember;range=0-4").size());
        checkUsers(objs.get(0).get("uniquemember;range=0-4"), 0, 4);
    }

    @Test
    public void testRequestUniqueMemberRange0_50() throws NamingException {
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=0-50");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=0-4"));
        Assert.assertEquals(5, objs.get(0).get("uniquemember;range=0-4").size());
        checkUsers(objs.get(0).get("uniquemember;range=0-4"), 0, 4);
    }

    @Test
    public void testRequestUniqueMemberRange1_3() throws NamingException {
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=1-3");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=1-3"));
        Assert.assertEquals(3, objs.get(0).get("uniquemember;range=1-3").size());
        checkUsers(objs.get(0).get("uniquemember;range=1-3"), 1, 3);
    }

    @Test
    public void testRequestUniqueMemberRange3_9() throws NamingException {
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=3-9");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=3-7"));
        Assert.assertEquals(5, objs.get(0).get("uniquemember;range=3-7").size());
        checkUsers(objs.get(0).get("uniquemember;range=3-7"), 3, 7);
    }

    @Test
    public void testRequestUniqueMemberRange8_() throws NamingException {
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=8-*");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=8-*"));
        Assert.assertEquals(3, objs.get(0).get("uniquemember;range=8-*").size());
        checkUsers(objs.get(0).get("uniquemember;range=8-*"), 8, 10);
    }

    @Test
    public void testRequestUniqueMemberRange1_() throws NamingException {
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=1-*");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=1-5"));
        Assert.assertEquals(5, objs.get(0).get("uniquemember;range=1-5").size());
        checkUsers(objs.get(0).get("uniquemember;range=1-5"), 1, 5);
    }

    @Test
    public void testRequestUniqueMemberRangeMinusMin() throws NamingException {
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=-10-*");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=0-4"));
        Assert.assertEquals(5, objs.get(0).get("uniquemember;range=0-4").size());
        checkUsers(objs.get(0).get("uniquemember;range=0-4"), 0, 4);
    }

    @Test
    public void testRequestUniqueMemberRange5_MinusMax() throws NamingException {
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=5-3");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=5-5"));
        Assert.assertEquals(1, objs.get(0).get("uniquemember;range=5-5").size());
        checkUsers(objs.get(0).get("uniquemember;range=5-5"), 5, 5);
    }

    @Test
    public void testRequestUniqueMemberRange15_18() throws NamingException {
        // AD returns an error => here just the last result is returned
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group1", SearchControls.SUBTREE_SCOPE, "uniquemember;range=15-18");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("uniquemember;range=10-*"));
        Assert.assertEquals(1, objs.get(0).get("uniquemember;range=10-*").size());
        checkUsers(objs.get(0).get("uniquemember;range=10-*"), 10, 10);
    }

    @Test
    public void testRequestMemberRange() throws NamingException {
        // other attribute is not ranged
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group3", SearchControls.SUBTREE_SCOPE, "member");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("member"));
        Assert.assertEquals(11, objs.get(0).get("member").size());
        checkUsers(objs.get(0).get("member"), 0, 10);
    }

    @Test
    public void testRequestMemberRangeExplicit() throws NamingException {
        // other attribute is not ranged even when explicit
        List<Map<String, List<String>>> objs = search("dc=sample,dc=com", "cn=group3", SearchControls.SUBTREE_SCOPE, "member;range=0-4");
        Assert.assertEquals(1, objs.size());
        Assert.assertNotNull(objs.get(0).get("member"));
        Assert.assertEquals(11, objs.get(0).get("member").size());
        checkUsers(objs.get(0).get("member"), 0, 10);
    }

    @Test
    public void testSearchAllGroups() throws NamingException {
        List<Map<String, List<String>>> objs = search("ou=Groups,dc=sample,dc=com", "(|(objectclass=groupofuniquenames)(objectclass=groupofnames))", SearchControls.ONELEVEL_SCOPE);
        Assert.assertEquals(3, objs.size());
        for (Map<String, List<String>> obj : objs) {
            Assert.assertNotNull(obj.get("cn"));
            Assert.assertEquals(1, obj.get("cn").size());
            String cn = obj.get("cn").get(0);
            switch (cn) {
                case "group1": // group1 should be ranged
                    Assert.assertNotNull(obj.get("uniquemember;range=0-4"));
                    Assert.assertEquals(5, obj.get("uniquemember;range=0-4").size());
                    checkUsers(obj.get("uniquemember;range=0-4"), 0, 4);
                    break;
                case "group2": // group2 has 3 members shold not be ranged
                    Assert.assertNotNull(obj.get("uniquemember"));
                    Assert.assertEquals(3, obj.get("uniquemember").size());
                    checkUsers(obj.get("uniquemember"), 0, 2);
                    break;
                case "group3": // group3 uses member so it should be not ranged
                    Assert.assertNotNull(obj.get("member"));
                    Assert.assertEquals(11, obj.get("member").size());
                    checkUsers(obj.get("member"), 0, 10);
                    break;
                default:
                    throw new IllegalStateException("Invalid group");
            }
        }
    }
}
